import json
from subprocess import run
from sys import stderr
from threading import Thread
from time import time, sleep
from typing import Tuple, Optional, Iterable, List

from nacl.bindings import crypto_sign_PUBLICKEYBYTES
from sqlalchemy import create_engine, LargeBinary, Column, REAL, BLOB, String, or_, and_, ForeignKeyConstraint, \
    Integer, UniqueConstraint, Index, Boolean
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, Query
from sqlalchemy.orm.exc import NoResultFound

from v25.messaging.encoding import NONCE_SIZE, Encoding
from v25.messaging.flags import Flags
from v25.messaging.message import Message
from v25.messaging.subject import Subject
from v25.storage.storage import EVENT_PUSH, EVENT_EDIT, EVENT_DELETE, PushStorage

Base = declarative_base()


class SSSJ(Base):
    """Subject Status Storage JSON"""
    __tablename__ = 'sssj'

    subject = Column(LargeBinary(crypto_sign_PUBLICKEYBYTES), primary_key=True)
    status = Column(String, nullable=False, default='{}')


class MsgSgn(Base):
    __tablename__ = 'msgsgns'

    sf = Column(LargeBinary(crypto_sign_PUBLICKEYBYTES), primary_key=True)
    st = Column(LargeBinary(crypto_sign_PUBLICKEYBYTES), primary_key=True)
    idn = Column(LargeBinary(NONCE_SIZE), primary_key=True)


class Msg(Base):
    __tablename__ = 'msgs'

    oid = Column(Integer, autoincrement=True, primary_key=True)
    sf = Column(LargeBinary(crypto_sign_PUBLICKEYBYTES), index=True)
    st = Column(LargeBinary(crypto_sign_PUBLICKEYBYTES), index=True)
    idn = Column(LargeBinary(NONCE_SIZE), index=True)
    ts = Column(REAL, nullable=False, index=True)
    en = Column(LargeBinary(NONCE_SIZE), nullable=False, index=True)
    ec = Column(BLOB, nullable=False)
    flags = Column(String, nullable=False)

    __table_args__ = (ForeignKeyConstraint((sf, st, idn,),
                                           (MsgSgn.sf, MsgSgn.st, MsgSgn.idn,),
                                           ),
                      UniqueConstraint(sf, st, idn, ),
                      Index("ix_msgs_full_en", sf, st, idn, en),
                      Index("ix_msgs_full_in", sf, st, idn, ),
                      Index("ix_msgs_full_ts", sf, st, ts, ),
                      Index("ix_msgs_full", sf, st, ),
                      )

    def sgn(self):
        return MsgSgn(sf=self.sf, st=self.st, idn=self.idn)

    def __repr__(self):
        return f"{type(self).__name__}(" \
               f"sf={self.sf!r}, st={self.st!r}, idn={self.idn!r}, ts={self.ts!r}," \
               f" en={self.en!r}, ec={self.ec!r}," \
               f" flags={self.flags!r}" \
               f")"

    @classmethod
    def from_message(cls, m: Message):
        return Msg(sf=m.sfrom.vkey.encode(), st=m.sto.vkey.encode(), idn=m.idnonce, ts=m.timestamp,
                   en=m.editnonce, ec=m.econtent,
                   flags=m.flags)

    def to_message(self):
        return Message(Subject(self.sf), Subject(self.st), self.idn, self.ts,
                       self.en, None, self.ec, self.flags)


class MsgEvent(Base):
    __tablename__ = 'msgevents'

    oid = Column(Integer, autoincrement=True, primary_key=True)
    sf = Column(LargeBinary(crypto_sign_PUBLICKEYBYTES), index=True)
    st = Column(LargeBinary(crypto_sign_PUBLICKEYBYTES), index=True)
    idn = Column(LargeBinary(NONCE_SIZE), index=True)
    ts = Column(REAL, nullable=False, index=True)
    en = Column(LargeBinary(NONCE_SIZE), nullable=False, index=True)

    __table_args__ = (ForeignKeyConstraint((sf, st, idn,),
                                           (MsgSgn.sf, MsgSgn.st, MsgSgn.idn,),
                                           ),
                      UniqueConstraint(sf, st, en, ),
                      Index("ix_msgevents_full_en", sf, st, en, ),
                      Index("ix_msgevents_full_ts", sf, st, ts, ),
                      Index("ix_msgevents_full", sf, st, ),
                      )

    @classmethod
    def from_message(cls, m: Message, code: int):
        return MsgEvent(sf=m.sfrom.vkey.encode(), st=m.sto.vkey.encode(), idn=m.idnonce, ts=time(),
                        en=bytes([code]) + Encoding.nonce())


class PushSub(Base):
    __tablename__ = 'pushsubs'

    subject = Column(LargeBinary(crypto_sign_PUBLICKEYBYTES), primary_key=True)
    subscription = Column(String, nullable=False)


class PushState(Base):
    __tablename__ = 'pushstate'

    subject = Column(LargeBinary(crypto_sign_PUBLICKEYBYTES), primary_key=True)
    notified = Column(Boolean, nullable=False)
    pushmsg = Column(Boolean, nullable=False)
    lastonline: Optional[float] = Column(REAL, nullable=True)

    @classmethod
    def of(cls, s: Subject, session) -> 'PushState':
        try:
            return session.query(PushState).filter_by(subject=s.vkey.encode()).one()
        except NoResultFound:
            state = PushState(subject=s.vkey.encode(), notified=False, pushmsg=False, lastonline=None)
            session.add(state)
            return state

    def afk_for(self, s: float):
        return self.lastonline is None or time() > self.lastonline + s

    def online(self):
        self.notified = False
        self.pushmsg = False
        self.lastonline = time()

    def notify(self) -> bool:
        if not self.pushmsg or self.notified or not self.afk_for(5):
            return False
        self.pushmsg = False
        self.notified = True
        return True


class DBStorage(PushStorage):
    def __init__(self, *args):
        self.args = args
        self.engine = create_engine(*args, echo=False)
        Base.metadata.create_all(self.engine)
        self.Session = sessionmaker(bind=self.engine)

    def check(self, subject: Subject) -> dict:
        session = self.Session()
        try:
            sssj = session.query(SSSJ).filter_by(subject=subject.vkey.encode()).one_or_none()
            status = json.loads(sssj.status) if sssj else {}
            if 'contacts' in status:
                query: Query = session.query(Msg).filter(or_(
                    Msg.sf == subject.vkey.encode(),
                    Msg.st == subject.vkey.encode(),
                ))
                query = query.with_entities(Msg.sf, Msg.st).distinct()
                contacts = set()
                for sf, st in query:
                    contacts.add(sf)
                    contacts.add(st)
                status['contacts'] = list(map(Encoding.encode, contacts))
            return status
        finally:
            session.close()

    def push(self, m: Message) -> None:
        session = self.Session()
        try:
            msg = Msg.from_message(m)
            session.add(msg)
            session.add(msg.sgn())
            self.event(session, m, EVENT_PUSH)
            session.commit()
            if m.sfrom == m.sto:
                return
            state = PushState.of(m.sto, session)
            state.pushmsg = True
            session.commit()
        finally:
            session.close()

    def edit(self, old: Message, new: Message):
        assert old.edited(new), 'edit misuse'
        session = self.Session()
        try:
            msg = self.one_alike(session, old)
            assert msg.en == old.editnonce, 'edit misuse'
            msgn = Msg.from_message(new)
            msg.en = msgn.en
            msg.ec = msgn.ec
            msg.flags = msgn.flags
            self.event(session, new, EVENT_EDIT)
            session.commit()
        finally:
            session.close()

    @staticmethod
    def one_alike(session, m: Message) -> Msg:
        return session.query(Msg).filter_by(sf=m.sfrom.vkey.encode(), st=m.sto.vkey.encode(), idn=m.idnonce).one()

    def delete(self, m: Message):
        session = self.Session()
        try:
            session.delete(self.one_alike(session, m))
            self.event(session, m, EVENT_DELETE)
            session.commit()
        finally:
            session.close()

    def pull(self, pair: Tuple[Subject, Subject], params: Optional[dict] = None) -> Iterable[Message]:
        if params is None:
            params = {}
        session = self.Session()
        try:
            cquery: Query = session.query(Msg).filter(or_(
                and_(
                    Msg.sf == pair[0].vkey.encode(),
                    Msg.st == pair[1].vkey.encode(),
                ),
                and_(
                    Msg.sf == pair[1].vkey.encode(),
                    Msg.st == pair[0].vkey.encode(),
                ),
            ))
            query: Query = cquery
            if 'ts' in params:
                if '>' in params:
                    query = query.filter(Msg.ts > params['ts']['>'])
                if '<' in params:
                    query = query.filter(Msg.ts < params['ts']['<'])
            if params.get('before'):
                query = query.filter(Msg.oid < cquery.filter(Msg.idn == Encoding.decode(params['before'])).one().oid)
            if params.get('after'):
                query = query.filter(Msg.oid > cquery.filter(Msg.idn == Encoding.decode(params['after'])).one().oid)
            if params.get('exact'):
                query = query.filter_by(idn=Encoding.decode(params['exact']))
            for flag in params.get('flags', ()):
                query = query.filter(Msg.flags.contains(flag))
            query = query.order_by(Msg.oid.desc())
            if 'limit' in params:
                query = query.limit(params['limit'])
            return map(Msg.to_message, list(query.from_self().order_by(Msg.oid)))
        finally:
            session.close()

    def flags(self, m: Message, flags: str):
        assert not Flags(flags).quable(), 'flags misuse'
        session = self.Session()
        try:
            msg: Msg = self.one_alike(session, m)
            assert msg.en == m.editnonce, 'flags misuse'
            assert Flags(msg.flags).quable(), 'flags misuse'
            msg.flags = flags
            session.commit()
        finally:
            session.close()

    def clearsssj(self):
        session = self.Session()
        try:
            session.query(SSSJ).delete()
            session.commit()
        finally:
            session.close()

    def ssssj(self, subject: str, status: str):
        """set SSSJ"""
        session = self.Session()
        try:
            subject = Subject(Encoding.decode(subject)).vkey.encode()
            session.query(SSSJ).filter_by(subject=subject).delete()
            session.add(SSSJ(subject=subject, status=status))
            session.commit()
        finally:
            session.close()

    @staticmethod
    def event(session, m: Message, code: int):
        if m.sfrom == m.sto:
            return
        session.add(MsgEvent.from_message(m, code))

    @staticmethod
    def ev_oid(query: Query, en: bytes) -> MsgEvent:
        return query.filter_by(en=en).one()

    def events(self, sfrom: Subject, sto: Subject, after: Optional[bytes]) -> Iterable[Tuple[bytes, bytes]]:
        session = self.Session()
        try:
            PushState.of(sto, session).online()
            session.commit()
            query: Query = session.query(MsgEvent)
            query = query.filter_by(sf=sfrom.vkey.encode(), st=sto.vkey.encode())
            if after is None:
                query = query.order_by(MsgEvent.oid.desc())
                query = query.limit(1)
            else:
                query = query.filter(MsgEvent.oid > self.ev_oid(query, after).oid)
                query = query.order_by(MsgEvent.oid)
            ev: MsgEvent
            return [(ev.en, ev.idn) for ev in query.all()]
        finally:
            session.close()

    def cleanevents(self):
        session = self.Session()
        try:
            session.query(MsgEvent).delete(MsgEvent.ts < time() - 604800)
            session.commit()
        finally:
            session.close()

    def subscribe(self, subject: Subject, subscription: dict):
        session = self.Session()
        try:
            subject = subject.vkey.encode()
            session.query(PushSub).filter_by(subject=subject).delete()
            session.add(PushSub(subject=subject, subscription=json.dumps(subscription)))
            session.commit()
        finally:
            session.close()

    def subscription(self, subject: Subject) -> Optional[dict]:
        session = self.Session()
        try:
            subscription: Optional[PushSub] = session.query(PushSub).filter_by(
                subject=subject.vkey.encode()).one_or_none()
            if subscription is None:
                return None
            return json.loads(subscription.subscription)
        finally:
            session.close()

    @staticmethod
    def pushsubscription(subscription: dict):
        assert run(["node", "v25pushx/push.js"], input=json.dumps({
            "subscription": subscription,
            "notification": {
                "notification": {
                    "title": "New Message (V25PUSH)"
                }
            }
        }), encoding='utf-8').returncode, "push subprocess failed"

    def pushpush(self, subject: Subject):
        session = self.Session()
        try:
            notify = PushState.of(subject, session).notify()
            session.commit()
        finally:
            session.close()
        if notify:
            subscription = self.subscription(subject)
            if subscription:
                self.pushsubscription(subscription)

    def push_once(self):
        session = self.Session()
        try:
            subs: List[PushSub] = session.query(PushSub).all()
        finally:
            session.close()
        for sub in subs:
            try:
                self.pushpush(Subject(sub.subject))
            except Exception as e:
                print('push failed', e, file=stderr)

    def _push_forever(self):
        while True:
            self.push_once()
            sleep(2)

    def _push_forever_run(self):
        storage = DBStorage(*self.args)
        storage._push_forever()

    def push_forever(self):
        Thread(target=self._push_forever_run, daemon=True).start()

    def pushing(self):
        self.push_forever()
        return self
